from datetime import datetime, timedelta

from celery import shared_task
from wbcore.contrib.authentication.models import User


@shared_task(queue="synchronization")
def google_webhook_resubscription() -> None:
    """
    A task to renew the google webhook subscriptions. The expiration date will be increased by 8 days.
    Only the subscriptions of users who still have a valid subscription will be renewed.
    """

    from .google import GoogleCalendarBackend

    user: User
    for user in User.objects.filter(metadata__google_backend__watch__isnull=False):
        GoogleCalendarBackend.stop_web_hook(user)
        user.refresh_from_db()
        new_timestamp_ms = round((datetime.now() + timedelta(days=8)).timestamp() * 1000)
        GoogleCalendarBackend.set_web_hook(user, new_timestamp_ms)



import random

import factory
from wbcore.contrib.directory.factories import EntryFactory

from wbcrm.models.groups import Group


class GroupFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Group

    title = factory.Faker("text", max_nb_chars=64)

    @factory.post_generation
    def members(self, create, extracted, **kwargs):
        if not create:
            return
        elif extracted:
            for member in extracted:
                self.members.add(member)
        else:
            for i in range(1, random.randrange(2, 10)):
                self.members.add(EntryFactory())



import pytest
from faker import Faker
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories.users import UserFactory
from wbcore.contrib.directory.models import Entry

from wbcrm.factories.accounts import AccountFactory, AccountRoleFactory
from wbcrm.models import Account
from wbcrm.viewsets.accounts import (
    AccountModelViewSet,
    AccountRepresentationViewSet,
    AccountRoleAccountModelViewSet,
    ChildAccountAccountModelViewSet,
    InheritedAccountRoleAccountModelViewSet,
)

fake = Faker()


@pytest.mark.django_db
class TestAccountViewset:
    @pytest.fixture
    def account_user(self):
        # True, we create a superuser

        if fake.pybool():
            user = UserFactory.create(is_superuser=True)
        else:
            user = UserFactory.create(is_superuser=True)
        entry = Entry.objects.get(id=user.profile.id)

        # Create a bunch of account and roles
        public_account = AccountFactory.create(is_public=True)
        AccountRoleFactory.create(account=public_account, entry=entry)
        AccountRoleFactory.create(account=public_account)
        private_account = AccountFactory.create(is_public=False)
        AccountRoleFactory.create(account=private_account, entry=entry)
        return user

    @pytest.mark.parametrize(
        "viewset_class",
        [
            AccountRepresentationViewSet,
            AccountModelViewSet,
        ],
    )
    def test_ensure_permission_on_account(self, account_user, viewset_class):
        request = APIRequestFactory().get("")
        request.user = account_user
        viewset = viewset_class(request=request)
        assert set(viewset.get_queryset()) == set(Account.objects.filter_for_user(account_user))

    def test_ensure_permission_for_nested_view(self, account_user, account_factory):
        request = APIRequestFactory().get("")
        request.user = account_user
        parent_account = Account.objects.first()
        account_factory.create(parent=parent_account, is_public=True)
        viewset = ChildAccountAccountModelViewSet(request=request, kwargs={"account_id": parent_account.id})

        assert set(viewset.get_queryset()) == set(
            Account.objects.filter(parent=parent_account).filter_for_user(account_user)
        )

    def test_ensure_permission_on_account_role_account(self, user, account_factory, account_role_factory):
        request = APIRequestFactory().get("")
        request.user = user
        parent_account = account_factory.create()
        child_account = account_factory.create(parent=parent_account)
        viewset_account_role = AccountRoleAccountModelViewSet(
            request=request, kwargs={"account_id": parent_account.id}
        )
        viewset_inherited_account_role = InheritedAccountRoleAccountModelViewSet(
            request=request, kwargs={"account_id": child_account.id}
        )
        with pytest.raises((AuthenticationFailed,)):
            viewset_account_role.get_queryset()
        with pytest.raises((AuthenticationFailed,)):
            viewset_inherited_account_role.get_queryset()

        role = account_role_factory.create(account=parent_account, entry=Entry.objects.get(id=user.profile.id))
        account_role_factory.create(account=parent_account)  # noise account role
        assert set(viewset_account_role.get_queryset()) == {role}

        account_role_factory.create(
            account=child_account, entry=Entry.objects.get(id=user.profile.id)
        )  # noise account role
        assert set(viewset_inherited_account_role.get_queryset()) == {role}



from typing import Type

from django.db.models import Q
from django.db.models.expressions import F
from django.db.models.query import QuerySet
from django.utils.translation import gettext_lazy as _
from wbcore import serializers
from wbcore.contrib.directory.models import Person
from wbcore.serializers.serializers import Serializer
from wbcrm.models import Activity, ActivityParticipant, ActivityType
from wbcrm.serializers import ActivityTypeRepresentationSerializer
from wbhuman_resources.models.kpi import KPI, KPIHandler
from wbhuman_resources.serializers import KPIModelSerializer


class NumberOfActivityKPISerializer(KPIModelSerializer):
    activity_type = serializers.PrimaryKeyRelatedField(
        required=True, many=True, queryset=ActivityType.objects.all(), label=_("Activity Type")
    )
    _activity_type = ActivityTypeRepresentationSerializer(source="activity_type", many=True)

    activity_area = serializers.ChoiceField(
        default="all",
        choices=[("only_internal", _("Only Internal")), ("only_external", _("Only External")), ("all", _("All"))],
    )
    person_participates = serializers.BooleanField(
        default=True,
        label=_("Participants of Activity"),
        help_text=_("Activities considered are related to the participants"),
    )
    person_created = serializers.BooleanField(
        default=True, label=_("Creator of activity"), help_text=_("Activities considered are related to the creator")
    )
    person_assigned = serializers.BooleanField(
        default=True,
        label=_("Person Assigned"),
        help_text=_("Activities considered are related to the persons assigned"),
    )

    def update(self, instance, validated_data):
        activity_type = validated_data.get(
            "activity_type",
            instance.additional_data["serializer_data"].get(
                "activity_type", list(ActivityType.objects.values_list("id", flat=True))
            ),
        )
        activity_area = validated_data.get(
            "activity_area",
            instance.additional_data["serializer_data"].get("activity_area", "all"),
        )

        person_participates = validated_data.get(
            "person_participates",
            instance.additional_data["serializer_data"].get("person_participates", True),
        )
        person_created = validated_data.get(
            "person_created",
            instance.additional_data["serializer_data"].get("person_created", True),
        )
        person_assigned = validated_data.get(
            "person_assigned",
            instance.additional_data["serializer_data"].get("person_assigned", True),
        )

        additional_data = instance.additional_data
        additional_data["serializer_data"]["activity_type"] = (
            [_type.id for _type in validated_data.get("activity_type")]
            if validated_data.get("activity_type")
            else activity_type
        )
        additional_data["serializer_data"]["activity_area"] = activity_area
        additional_data["serializer_data"]["person_participates"] = person_participates
        additional_data["serializer_data"]["person_created"] = person_created
        additional_data["serializer_data"]["person_assigned"] = person_assigned

        additional_data["list_data"] = instance.get_handler().get_list_data(additional_data["serializer_data"])
        validated_data["additional_data"] = additional_data

        return super().update(instance, validated_data)

    class Meta(KPIModelSerializer.Meta):
        fields = (
            *KPIModelSerializer.Meta.fields,
            "activity_type",
            "_activity_type",
            "activity_area",
            "person_participates",
            "person_created",
            "person_assigned",
        )


class NumberOfActivityKPI(KPIHandler):
    def get_name(self) -> str:
        return _("Number of Activities")

    def get_serializer(self) -> Type[Serializer]:
        return NumberOfActivityKPISerializer

    def annotate_parameters(self, queryset: QuerySet[KPI]) -> QuerySet[KPI]:
        return queryset.annotate(
            activity_type=F("additional_data__serializer_data__activity_type"),
            activity_area=F("additional_data__serializer_data__activity_area"),
            person_participates=F("additional_data__serializer_data__person_participates"),
            person_created=F("additional_data__serializer_data__person_created"),
            person_assigned=F("additional_data__serializer_data__person_assigned"),
        )

    def get_list_data(self, serializer_data: dict) -> list[str]:
        activity_types = list(
            ActivityType.objects.filter(pk__in=serializer_data["activity_type"]).values_list("title", flat=True)
        )
        return [
            _("Activity Type: {types}").format(types=activity_types),
            _("Activity Area: {area}").format(area=serializer_data["activity_area"]),
            _("Person Participates: {participating}").format(participating={serializer_data["person_participates"]}),
            _("Person Created: {created}").format(created=serializer_data["person_created"]),
            _("Person Assigned: {assigned}").format(assigned=serializer_data["person_assigned"]),
        ]

    def get_display_grid(self) -> list[list[str]]:
        return [
            ["activity_type"] * 3,
            ["activity_area"] * 3,
            ["person_created", "person_assigned", "person_participates"],
        ]

    def evaluate(self, kpi: "KPI", evaluated_person=None, evaluation_date=None) -> int:
        persons = (
            [evaluated_person.id] if evaluated_person else kpi.evaluated_persons.all().values_list("id", flat=True)
        )
        serializer_data = kpi.additional_data.get("serializer_data")
        qs = Activity.objects.filter(
            period__startswith__date__gte=kpi.period.lower,
            period__endswith__date__lte=evaluation_date if evaluation_date else kpi.period.upper,
        ).exclude(status=Activity.Status.CANCELLED)
        qs = (
            qs.filter(type__id__in=serializer_data.get("activity_type"))
            if serializer_data.get("activity_type")
            else qs
        )

        condition = None
        if serializer_data.get("person_created") or serializer_data.get("person_created") is None:
            condition = Q(creator__in=persons)
        if serializer_data.get("person_assigned") or serializer_data.get("person_assigned") is None:
            condition = (condition | Q(assigned_to__in=persons)) if condition else Q(assigned_to__in=persons)
        if serializer_data.get("person_participates") or serializer_data.get("person_participates") is None:
            condition = (condition | Q(participants__in=persons)) if condition else Q(participants__in=persons)
        if condition:
            qs = qs.filter(condition)

        if (activity_area := serializer_data.get("activity_area")) and (activity_area != "all"):
            participant_ids = set(
                ActivityParticipant.objects.filter(activity__id__in=qs.values_list("id", flat=True)).values_list(
                    "participant", flat=True
                )
            )
            employee_ids = Person.objects.filter_only_internal().values_list("id", flat=True)
            externals = set(
                Person.objects.filter(Q(id__in=participant_ids) & ~Q(id__in=employee_ids)).values_list("id", flat=True)
            )
            if activity_area == "only_internal":
                qs = qs.exclude(
                    Q(creator__in=externals) | Q(assigned_to__in=externals) | Q(participants__id__in=externals)
                )
            elif activity_area == "only_external":
                qs = qs.filter(
                    Q(creator__in=externals) | Q(assigned_to__in=externals) | Q(participants__id__in=externals)
                )
        return qs.distinct("id").count()



from unittest.mock import patch

import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.workflow.factories import (
    ProcessStepFactory,
    RandomChildStepFactory,
    UserStepFactory,
)

from wbcrm.workflows.assignee_methods import activity_assignee


@pytest.mark.django_db
class TestAssigneeMethods:
    api_factory = APIRequestFactory()

    def test_activity_assignee(self, activity_factory):
        user = UserFactory()
        activity = activity_factory(assigned_to=user.profile)
        process_step = ProcessStepFactory(process__instance=activity)
        assert activity_assignee(process_step) == user

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    def test_activity_assignee_no_activity(self, mock_failed):
        step = RandomChildStepFactory(exclude_factories=[UserStepFactory])
        process_step = ProcessStepFactory(step=step)
        assert activity_assignee(process_step) is None
        assert mock_failed.call_args.args[0] == process_step

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    def test_activity_assignee_no_user_account(self, mock_failed, activity_factory):
        assignee = PersonFactory()
        activity = activity_factory(assigned_to=assignee)
        step = RandomChildStepFactory(exclude_factories=[UserStepFactory])
        process_step = ProcessStepFactory(step=step, process__instance=activity)
        assert activity_assignee(process_step) is None
        assert mock_failed.call_args.args[0] == process_step



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

# Representations
router.register(r"activityrepresentation", viewsets.ActivityRepresentationViewSet, basename="activityrepresentation")
router.register(r"grouprepresentation", viewsets.GroupRepresentationViewSet, basename="grouprepresentation")
router.register(r"group", viewsets.GroupModelViewSet, basename="group")
router.register(r"activitytype", viewsets.ActivityTypeModelViewSet, basename="activitytype")
router.register(
    r"activitytyperepresentation", viewsets.ActivityTypeRepresentationViewSet, basename="activitytyperepresentation"
)

router.register(r"product", viewsets.ProductModelViewSet, basename="product")
router.register(
    r"productrepresentation",
    viewsets.ProductRepresentationViewSet,
    basename="productrepresentation",
)


# Activity
router.register(r"activity", viewsets.ActivityViewSet, basename="activity")
# used to create new activity instances

router.register(r"activitychart", viewsets.ActivityChartModelViewSet, basename="activitychart")


activity_router = WBCoreRouter()
activity_router.register(
    r"activity-participant",
    viewsets.ActivityParticipantModelViewSet,
    basename="activity-participant",
)

company_router = WBCoreRouter()
company_router.register(
    r"company-interestedproduct",
    viewsets.ProductCompanyRelationshipCompanyModelViewSet,
    basename="company-interestedproduct",
)

router.register(r"account", viewsets.AccountModelViewSet, basename="account")
router.register(r"accountrepresentation", viewsets.AccountRepresentationViewSet, basename="accountrepresentation")
router.register(
    r"accountroletyperepresentation",
    viewsets.AccountRoleTypeRepresentationViewSet,
    basename="accountroletyperepresentation",
)

account_router = WBCoreRouter()
account_router.register(r"childaccount", viewsets.ChildAccountAccountModelViewSet, basename="account-childaccount")
account_router.register(r"accountrole", viewsets.AccountRoleAccountModelViewSet, basename="account-accountrole")
account_router.register(
    r"inheritedrole", viewsets.InheritedAccountRoleAccountModelViewSet, basename="account-inheritedrole"
)

urlpatterns = [
    path("", include(router.urls)),
    path("activity/<int:activity_id>/", include(activity_router.urls)),
    path("company/<int:company_id>/", include(company_router.urls)),
    path("sync/", include(("wbcrm.synchronization.urls", "sync"), namespace="sync")),
    path("account/<int:account_id>/", include(account_router.urls)),
]



# import datetime
# import json
# from unittest.mock import patch

# import pytest
# from django.utils import timezone
# from dynamic_preferences.registries import global_preferences_registry
# from rest_framework.test import APIRequestFactory
# from wbcore.test.utils import get_or_create_superuser
# from wbcrm.models import ActivityParticipant

# from ..google_calendar_backend import GoogleCalendarBackend
# from .test_data import credentials, event_data, execute_service, service_data


# class TestInitialisationGoogleCalendarBackend:
#     @pytest.fixture()
#     def activity_with_user(self, activity_factory):
#         user = get_or_create_superuser()
#         user.metadata = {"google_backend": {"watch": {"id": "fake_id"}}}
#         user.save()
#         return activity_factory(creator=user.profile)

#     @pytest.fixture()
#     def activity_with_user2(self, activity_factory):
#         user = get_or_create_superuser()
#         user.metadata = {"google_backend": {"watch": {"id": "fake_id2"}}}
#         user.save()
#         return activity_factory(creator=user.profile)

#     @pytest.mark.parametrize("credentials", ["", credentials])
#     def test_get_service_account_file(self, credentials):
#         from ..google_calendar_backend import GoogleCalendarBackend

#         global_preferences_registry.manager()["wbactivity_sync__google_sync_credentials"] = credentials

#         backend = GoogleCalendarBackend()

#         if not credentials:
#             with pytest.raises(ValueError) as excinfo:
#                 file = backend._get_service_account_file()
#                 assert (
#                     "The Google credentials are not set. You cannot use the Google Calendar Backend without the Google credentials."
#                     in str(excinfo.value)
#                 )
#         else:
#             file = backend._get_service_account_file()
#             assert file == json.loads(credentials)

#     @pytest.mark.parametrize("credentials", [credentials])
#     def test_get_service_account_url(self, credentials):
#         from ..google_calendar_backend import GoogleCalendarBackend

#         global_preferences_registry.manager()["wbactivity_sync__google_sync_credentials"] = credentials
#         assert GoogleCalendarBackend()._get_service_account_url()

#     def test_get_service_user_email(self, activity_with_user, activity_with_user2, activity_factory):
#         global_preferences_registry.manager()["wbcrm__main_company"] = employee.employers.first().id

#         activity = activity_factory()
#         activity1 = activity_factory()
#         activity2 = activity_with_user2
#         activity3 = activity_with_user
#         activity4 = activity_factory(participants=(employee,))
#         tomorrow = (timezone.now() + datetime.timedelta(days=1)).replace(tzinfo=None)
#         activity3.creator.user_account.metadata["google_backend"]["watch"]["expiration"] = (
#             datetime.datetime.timestamp(tomorrow) * 1000
#         )
#         activity3.creator.user_account.save()
#         employee.user_account.metadata = activity3.creator.user_account.metadata
#         employee.user_account.save()
#         backend = GoogleCalendarBackend()
#         assert backend._get_service_user_email(activity) == ""
#         assert backend._get_service_user_email(activity1) == ""
#         assert backend._get_service_user_email(activity2) == ""
#         assert backend._get_service_user_email(activity3) == activity3.creator.user_account.email
#         assert backend._get_service_user_email(activity4)

#     @patch("wbcrm.synchronization.activity.backends.google.google_calendar_backend.Credentials")
#     @patch("wbcrm.synchronization.activity.backends.google.google_calendar_backend.build")
#     @pytest.mark.parametrize("credentials", ["", credentials])
#     def test_build_service(self, mock_build, mock_credentials, credentials):
#         global_preferences_registry.manager()["wbactivity_sync__google_sync_credentials"] = credentials
#         user = get_or_create_superuser()
#         backend = GoogleCalendarBackend()
#         if not credentials:
#             with pytest.raises(ValueError) as excinfo:
#                 resource = backend._build_service(user.email)
#                 assert backend.error_messages["service_build_error"] in str(excinfo.value)
#         else:
#             resource = backend._build_service(user.email)
#             assert resource


# @patch("wbcrm.synchronization.activity.backends.google.google_calendar_backend.GoogleCalendarBackend._build_service")
# @patch(
#     "wbcrm.synchronization.activity.backends.google.google_calendar_backend.GoogleCalendarBackend._get_service_user_email"
# )
# @pytest.mark.parametrize(
#     "sync_past_activity, event", [(False, ""), (True, ""), (False, event_data), (True, event_data)]
# )
# @pytest.mark.django_db
# class TestGoogleCalendarBackend:
#     @pytest.fixture()
#     def activity_fixture(self, activity_factory, event):
#         user = get_or_create_superuser()
#         user.metadata = {"google_backend": {"watch": {"id": "fake_id"}}}
#         user.save()
#         if event:
#             metadata = {"google_backend": {"event": event}}
#             return activity_factory(creator=user.profile, metadata=metadata, external_id=event["id"])
#         else:
#             metadata = {}
#             return activity_factory(creator=user.profile, metadata=metadata, external_id=None)

#     def test_create_external_activity(
#         self, mock_user_email, mock_service, sync_past_activity, event, activity_factory
#     ):
#         global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"] = sync_past_activity
#         now = timezone.now()
#         yesterday = now - datetime.timedelta(days=1)
#         tomorrow = now + datetime.timedelta(days=1)

#         activity_passed = activity_factory(start=yesterday, end=yesterday + datetime.timedelta(hours=1))
#         activity = activity_factory(start=tomorrow, end=tomorrow + datetime.timedelta(hours=1))
#         activity_with_parent = activity_factory(
#             start=tomorrow, end=tomorrow + datetime.timedelta(hours=1), parent_occurrence=activity_passed
#         )

#         mock_service.return_value = service_data()
#         with patch.object(execute_service, "execute", return_value=event):

#             backend = GoogleCalendarBackend()
#             if event:
#                 backend.create_external_activity(activity_passed)
#                 backend.create_external_activity(activity)
#                 backend.create_external_activity(activity_with_parent)
#                 assert (
#                     mock_service.call_count == 2
#                 )  # 2 instead of 3 because an activity with a parent cannot be synchronized

#             else:
#                 with pytest.raises(ValueError) as excinfo:
#                     backend.create_external_activity(activity_passed)
#                     backend.create_external_activity(activity)
#                     backend.create_external_activity(activity_with_parent)
#                     assert "Could not create the external google event. Exception" in str(excinfo.value)

#                     if sync_past_activity:
#                         assert mock_service.call_count == 1
#                     else:
#                         assert mock_service.call_count == 1
#             activity_passed.refresh_from_db()
#             activity.refresh_from_db()
#             activity_with_parent.refresh_from_db()

#             assert activity_with_parent.metadata == {}
#             assert activity_with_parent.external_id is None

#             if event:
#                 assert activity.metadata
#                 assert activity.external_id == event["id"]
#             else:
#                 assert activity.metadata == {}
#                 assert activity.external_id is None

#             if sync_past_activity and event:
#                 assert activity_passed.metadata
#                 assert activity_passed.external_id == event["id"]
#             else:
#                 assert activity_passed.metadata == {}
#                 assert activity_passed.external_id is None

#     def test_delete_external_activity(
#         self, mock_user_email, mock_service, sync_past_activity, event, activity_factory
#     ):
#         global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"] = sync_past_activity
#         now = timezone.now()
#         yesterday = now - datetime.timedelta(days=1)
#         tomorrow = now + datetime.timedelta(days=1)

#         activity_passed = activity_factory(start=yesterday, end=yesterday + datetime.timedelta(hours=1))
#         activity = activity_factory(start=tomorrow, end=tomorrow + datetime.timedelta(hours=1))
#         activity_with_parent = activity_factory(
#             start=tomorrow, end=tomorrow + datetime.timedelta(hours=1), parent_occurrence=activity_passed
#         )

#         mock_service.return_value = service_data()
#         with patch.object(execute_service, "execute", return_value=None):

#             backend = GoogleCalendarBackend()
#             backend.delete_external_activity(activity_passed)
#             backend.delete_external_activity(activity)
#             backend.delete_external_activity(activity_with_parent)
#             assert mock_service.call_count == 3

#     def test_update_external_activity(
#         self, mock_user_email, mock_service, sync_past_activity, event, activity_factory
#     ):
#         global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"] = sync_past_activity
#         now = timezone.now()
#         yesterday = now - datetime.timedelta(days=1)
#         tomorrow = now + datetime.timedelta(days=1)

#         metadata = {"google_backend": {"instance": event if event else {}}}
#         external_id = event["id"] if event else None
#         activity_no_metadata = activity_factory(start=tomorrow, end=tomorrow + datetime.timedelta(hours=1))
#         activity_passed = activity_factory(
#             start=yesterday, end=yesterday + datetime.timedelta(hours=1), external_id=external_id, metadata=metadata
#         )
#         activity = activity_factory(
#             start=tomorrow, end=tomorrow + datetime.timedelta(hours=1), external_id=external_id, metadata=metadata
#         )
#         activity_with_parent = activity_factory(
#             start=tomorrow,
#             end=tomorrow + datetime.timedelta(hours=1),
#             parent_occurrence=activity_passed,
#             external_id=external_id,
#             metadata=metadata,
#         )

#         mock_service.return_value = service_data()
#         with patch.object(execute_service, "execute", return_value=event if event else {"id": None}):
#             backend = GoogleCalendarBackend()
#             backend.update_external_activity(activity_no_metadata)
#             backend.update_external_activity(activity_passed)
#             backend.update_external_activity(activity)
#             backend.update_external_activity(activity_with_parent)

#             assert (
#                 mock_service.call_count == 5
#             )  # rather than 4 because we call create_external_activity if no metadata
#             assert activity_no_metadata.external_id == external_id
#             assert activity_no_metadata.metadata == {"google_backend": {"event": event if event else {"id": None}}}

#     @pytest.mark.parametrize("response_status", [*list(ActivityParticipant.ParticipationStatus)])
#     def test_send_participant_response_external_activity(
#         self,
#         mock_user_email,
#         mock_service,
#         response_status,
#         sync_past_activity,
#         event,
#         person_factory,
#         activity_fixture,
#     ):
#         global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"] = sync_past_activity
#         activity = activity_fixture
#         person = person_factory()
#         activity.participants.add(person, activity.creator)
#         activity.save()
#         creator_participant = ActivityParticipant.objects.filter(
#             activity=activity, participant=activity.creator
#         ).first()
#         mock_service.return_value = service_data()
#         with patch.object(execute_service, "execute", return_value=event if event else {"id": None}) as mock_execute:
#             backend = GoogleCalendarBackend()
#             activityparticipant = activity.activity_participants.filter(participant=person).first()
#             backend.send_participant_response_external_activity(activityparticipant, response_status)
#             backend.send_participant_response_external_activity(creator_participant, response_status)
#             if event:
#                 assert mock_execute.call_count
#             else:
#                 assert mock_execute.call_count == 0

#     @patch(
#         "wbcrm.synchronization.activity.backends.google.google_calendar_backend.GoogleCalendarBackend.sync_with_external_calendar"
#     )
#     def test_sync_with_external_calendar(
#         self, mock_handle_changes, mock_user_email, mock_service, sync_past_activity, event
#     ):
#         user = get_or_create_superuser()
#         request = APIRequestFactory().get("", **{"HTTP_X-Goog-Channel-Id": user.id})
#         backend = GoogleCalendarBackend()
#         backend.sync_with_external_calendar(request)
#         assert mock_handle_changes.call_count == 1

#     def test_get_sync_token(self, mock_user_email, mock_service, sync_past_activity, event):
#         user = get_or_create_superuser()
#         user.metadata = {"google_backend": {"sync_token": "fake_token"}}
#         user.save()
#         backend = GoogleCalendarBackend()
#         assert backend.get_sync_token(user) == "fake_token"

#     def test_get_external_activity(self, mock_user_email, mock_service, sync_past_activity, event):
#         pass

#     def test_forward_external_activity(self, mock_user_email, mock_service, sync_past_activity, event):
#         pass

#     @patch(
#         "wbcrm.synchronization.activity.backends.google.google_calendar_backend.GoogleCalendarBackend._get_service_account_url"
#     )
#     def test_set_web_hook(self, mock_url, mock_user_email, mock_service, sync_past_activity, event):
#         user = get_or_create_superuser()
#         watch = {"id": "fake_id"}
#         mock_service.return_value = service_data()
#         with patch.object(execute_service, "execute", return_value=watch):
#             GoogleCalendarBackend().set_web_hook(user)
#         assert user.metadata["google_backend"]["watch"] == watch

#     def test_stop_web_hook(self, mock_user_email, mock_service, sync_past_activity, event):
#         user = get_or_create_superuser()
#         user.metadata = {"google_backend": {"watch": {"id": "fake_id", "resourceId": "fake_resource_id"}}}
#         user.save()
#         watch = {"id": "fake_id"}
#         mock_service.return_value = service_data()
#         with patch.object(execute_service, "execute", return_value=watch):
#             GoogleCalendarBackend().stop_web_hook(user)
#         assert user.metadata["google_backend"] == {}

#     def test_check_web_hook(self, mock_user_email, mock_service, sync_past_activity, event):
#         tomorrow = (timezone.now() + datetime.timedelta(days=1)).replace(tzinfo=None)
#         user = get_or_create_superuser()
#         user.metadata = {"google_backend": {"watch": {"expiration": datetime.datetime.timestamp(tomorrow) * 1000}}}
#         user.save()
#         with pytest.warns() as record:
#             GoogleCalendarBackend().check_web_hook(user)
#         assert len(record) == 1



from dynamic_preferences.registries import global_preferences_registry
from wbcore.utils.importlib import import_from_dotted_path


def get_backend():
    from wbcrm.synchronization.activity.controller import ActivityController

    if backend := global_preferences_registry.manager()["wbactivity_sync__sync_backend_calendar"]:
        backend = import_from_dotted_path(backend)
        return ActivityController(backend=backend)



from django.contrib import admin
from mptt.admin import MPTTModelAdmin

from wbcrm.models.accounts import (
    Account,
    AccountRole,
    AccountRoleType,
    AccountRoleValidity,
)


class AccountRoleValidityTabularInline(admin.TabularInline):
    model = AccountRoleValidity
    fields = ("timespan",)
    extra = 0


class AccountRoleTabularInline(admin.TabularInline):
    model = AccountRole
    fields = ("role_type", "entry", "weighting", "is_hidden")
    autocomplete_fields = ["entry", "account"]
    extra = 0
    show_change_link = True


class AccountTabularInline(admin.TabularInline):
    model = Account
    fields = ("title", "status", "owner", "is_terminal_account", "is_active", "is_public")
    autocomplete_fields = ["owner", "parent"]
    fk_name = "parent"
    show_change_link = True
    extra = 0


@admin.register(AccountRoleType)
class AccountRoleTypeModelAdmin(admin.ModelAdmin):
    list_display = ("title", "key")
    search_fields = ("title", "key")


@admin.register(AccountRole)
class AccountRoleModelAdmin(admin.ModelAdmin):
    list_display = ("role_type", "entry", "account", "is_hidden")
    inlines = [AccountRoleValidityTabularInline]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("entry", "account")


@admin.register(Account)
class AccountModelAdmin(MPTTModelAdmin):
    mptt_level_indent = 20
    fsm_field = ["status"]
    search_fields = ["title"]
    list_display = ["computed_str", "status", "is_active", "is_terminal_account", "is_public", "owner"]
    inlines = [AccountRoleTabularInline, AccountTabularInline]
    autocomplete_fields = ["owner", "parent"]

    def get_queryset(self, request):
        return Account.all_objects.select_related("owner", "parent")



import json
from contextlib import suppress
from datetime import date, timedelta
from urllib.parse import urlparse

import pandas as pd
import requests
from django.contrib.sites.models import Site
from django.db.utils import ProgrammingError
from django.http import HttpResponse
from django.utils import timezone
from dynamic_preferences.registries import global_preferences_registry
from rest_framework import status

from .parser import parse


class MicrosoftGraphAPI:
    def __init__(self):
        with suppress(ProgrammingError):
            global_preferences = global_preferences_registry.manager()
            if (credentials := global_preferences["wbactivity_sync__outlook_sync_credentials"]) and (
                serivce_account_file := json.loads(credentials)
            ):
                self.authority = serivce_account_file.get("authority")
                self.client_id = serivce_account_file.get("client_id")
                self.client_secret = serivce_account_file.get("client_secret")
                self.token_endpoint = serivce_account_file.get("token_endpoint")
                self.notification_url = serivce_account_file.get("notification_url")
                self.graph_url = serivce_account_file.get("graph_url")

                if global_preferences["wbactivity_sync__outlook_sync_access_token"] == "":
                    self._get_access_token()
                else:
                    # try to get a data in the api (i.e list of applications), update the token if it is expired
                    self.applications()
            else:
                self.authority = None
                self.client_id = None
                self.client_secret = None
                self.token_endpoint = None
                self.notification_url = None
                self.graph_url = None

    def _get_administrator_consent(self) -> HttpResponse:
        url = f"{self.authority}/adminconsent?client_id={self.client_id}&state=12345&redirect_uri={self.notification_url}"
        if urlparse(url):
            response = self._query(url, access_token=False)
            if response:
                return response
            else:
                raise ValueError("get administrator consent does not return response 200")
        else:
            raise ValueError("Invalid URL")

    def _get_access_token(self) -> str | None:
        # Get administrator consent
        self._get_administrator_consent()
        # Get an access token
        url = f"{self.authority}{self.token_endpoint}"
        payload = {
            "grant_type": "client_credentials",
            "client_id": self.client_id,
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": self.client_secret,
        }

        response = self._query(url, method="POST", data=payload, is_json=False, access_token=False)
        data = None
        if response:
            if response.json():
                data = response.json().get("access_token")
                global_preferences = global_preferences_registry.manager()
                global_preferences["wbactivity_sync__outlook_sync_access_token"] = data
        else:
            raise ValueError(response, response.json())
        return data

    def _subscribe(
        self, resource: str, change_type: str, minutes: int = 4230, raise_error: bool = False
    ) -> dict | None:
        subscription_data: dict = {
            "changeType": change_type,
            "notificationUrl": self.notification_url,
            "resource": resource,
            "clientState": global_preferences_registry.manager()["wbactivity_sync__outlook_sync_client_state"],
            "latestSupportedTlsVersion": "v1_2",
        }
        if minutes:
            # maximum time of subscription 4230 minutes (under 3 days)
            date = timezone.now() + timedelta(minutes=minutes)
            date = date.strftime("%Y-%m-%dT%H:%M:%SZ")
            subscription_data["expirationDateTime"] = date
        url = f"{self.graph_url}/subscriptions"
        response = self._query(url, method="POST", data=json.dumps(subscription_data))
        data = None
        if response and response.status_code == status.HTTP_201_CREATED:
            if response.json():
                data = parse(response.json(), scalar_value=True)
                if data and isinstance(data, list):
                    data = data[0]
        elif raise_error:
            raise ValueError(response, response.json())
        return data

    def _unsubscribe(self, subscription_id: str) -> HttpResponse:
        url = f"{self.graph_url}/subscriptions/{subscription_id}"
        return self._query(url, method="DELETE")

    def _renew_subscription(self, subscription_id: str, minutes: int = 4230, raise_error: bool = False) -> dict | None:
        url = f"{self.graph_url}/subscriptions/{subscription_id}"
        date = timezone.now() + timedelta(minutes=minutes)
        date = date.strftime("%Y-%m-%dT%H:%M:%SZ")
        data = {"expirationDateTime": date}
        response = self._query(url, method="PATCH", data=json.dumps(data))
        data = None
        if response and response.status_code == status.HTTP_200_OK:
            if response.json():
                data = parse(response.json(), scalar_value=True)
                if data and isinstance(data, list):
                    data = data[0]
        elif raise_error:
            raise ValueError(response, response.json())
        return data

    def subscriptions(self) -> list[dict]:
        url = f"{self.graph_url}/subscriptions"
        response = self._query(url)
        data = []
        if response:
            if datum := response.json():
                data = parse(datum.get("value"))
                url = datum.get("@odata.nextLink")
                while url:
                    response = self._query(url)
                    if datum := response.json():
                        data += parse(datum.get("value"))
                        url = datum.get("@odata.nextLink")
                    else:
                        url = None
        else:
            raise ValueError(response, response.json())
        return data

    def subscription(self, subscription_id: str, raise_error: bool = False) -> dict | None:
        url = f"{self.graph_url}/subscriptions/{subscription_id}"
        response = self._query(url)
        data = None
        if response and response.status_code == status.HTTP_200_OK:
            if response.json():
                data = parse(response.json(), scalar_value=True)
                if data and isinstance(data, list):
                    data = data[0]
        elif raise_error:
            raise ValueError(response, response.json())
        return data

    def applications(self) -> dict | None:
        # List of applications in MS graph
        url = f"{self.graph_url}/applications"
        response = self._query(url)
        data = None
        if response:
            if datum := response.json():
                data = parse(datum.get("value"))
                url = datum.get("@odata.nextLink")
                while url:
                    response = self._query(url)
                    if datum := response.json():
                        data += parse(datum.get("value"))
                        url = datum.get("@odata.nextLink")
                    else:
                        url = None
        elif response.status_code == status.HTTP_401_UNAUTHORIZED:
            self._get_access_token()
            if response := self._query(url):
                data = response.json()
        else:
            raise ValueError(response, response.json())
        return data

    def user(self, email: str, raise_error: bool = False) -> dict | None:
        query_params = {"$select": "id, userPrincipalName, displayName"}
        url = f"{self.graph_url}/users/{email}"
        response = self._query(url, params=query_params)
        data = None
        if response:
            if response.json():
                data = parse(response.json(), scalar_value=True)
                if data and isinstance(data, list):
                    data = data[0]
        elif raise_error:
            raise ValueError(response, response.json())
        return data

    def users(self, filter_params: bool = True) -> dict:
        query_params = {
            "$select": "id,displayName,businessPhones,mobilePhone, userPrincipalName, mail,email, mailNickname, givenName, surname, imAddresses"
        }
        url = f"{self.graph_url}/users"
        if filter_params:
            response = self._query(url, params=query_params)
        else:
            response = self._query(url)
        data = None
        if response:
            data = parse(response.json().get("value"))
        else:
            raise ValueError(response, response.json())
        return data

    def get_tenant_id(self, email: str) -> str | None:
        try:
            if msuser := self.user(email):
                return msuser.get("id", None)
        except Exception:
            pass
        return None

    def delta_changes_events(self, tenant_id: str, minutes: int) -> list:
        start = (timezone.now() - timedelta(minutes=minutes)).strftime("%Y-%m-%dT%H:%M:%SZ")
        end = (timezone.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
        url = f"{self.graph_url}/users/{tenant_id}/calendarView/delta?startdatetime={start}&enddatetime={end}"
        response = self._query(url)
        datum = []
        if response:
            if datum := response.json():
                _value = datum.get("value")
                external_event_list = parse(_value)
                url = datum.get("@odata.deltaLink")
                while _value:
                    response = self._query(url)
                    datum = response.json()
                    _value = datum.get("value")
                    external_event_list += parse(_value)
                    url = datum.get("@odata.deltaLink")
            else:
                raise ValueError(response, response.json())
        return datum

    def create_event(self, tenant_id: str, event_body: dict) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events"
        return self._query_create(url, event_body)

    def delete_event(self, tenant_id: str, external_id: str) -> None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}"
        return self._query(url, method="DELETE")

    def update_event(self, tenant_id: str, external_id: str, event_body: dict) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}"
        return self._query_update(url, event_body)

    def get_event(
        self, tenant_id: str, external_id: str, extension: bool = False, extension_id: str = ""
    ) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}"
        if extension:
            extension_id = (
                extension_id if extension_id else ".".join(reversed(Site.objects.get_current().domain.split(".")))
            )
            url += f"?$expand=Extensions($filter=Id eq '{extension_id}')"
        return self._query_get(url)

    def update_or_create_extension_event(
        self, tenant_id: str, external_id: str, extension_body: dict, extension_id: str = ""
    ):
        extension_id = (
            extension_id if extension_id else ".".join(reversed(Site.objects.get_current().domain.split(".")))
        )
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/extensions"
        data = {"@odata.type": "microsoft.graph.openTypeExtension", "extensionName": extension_id, **extension_body}
        if self.get_extension_event(tenant_id, external_id, extension_id):
            url += f"/{extension_id}"
            return self._query_update(url, data)
        else:
            return self._query_create(url, data)

    def get_extension_event(self, tenant_id: str, external_id: str, extension_id: str = "") -> dict | None:
        extension_id = (
            extension_id if extension_id else ".".join(reversed(Site.objects.get_current().domain.split(".")))
        )
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/extensions/{extension_id}"
        return self._query_get(url)

    def delete_extension_event(self, tenant_id: str, external_id: str, extension_id: str = "") -> None:
        extension_id = (
            extension_id if extension_id else ".".join(reversed(Site.objects.get_current().domain.split(".")))
        )
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/extensions/{extension_id}"
        return self._query(url, method="DELETE")

    def forward_event(self, tenant_id: str, external_id: str, participants: list) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/forward"
        data = {
            "ToRecipients": participants,
        }
        return self._query_create(url, data)

    def get_event_by_uid(self, tenant_id: str, uid: str) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events?$filter=uid eq '{uid}'"
        event = list_data[0] if (list_data := self._query_get_list(url)) else None
        return event

    def get_list_events(self, tenant_id: str) -> list:
        url = f"{self.graph_url}/users/{tenant_id}/events"
        return self._query_get_list(url)

    def get_instances_event(self, tenant_id: str, external_id: str, start: str | date, end: str | date) -> list:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/instances?startDateTime={start}&endDateTime={end}"
        list_data = self._query_get_list(url)
        return list_data if list_data else []

    def get_instances_event_by_resource(self, resource: str, start: str | date, end: str | date) -> list:
        url = f"{self.graph_url}/{resource}/instances?startDateTime={start}&endDateTime={end}"
        list_data = self._query_get_list(url)
        return list_data if list_data else []

    def get_event_by_resource(self, resource: str, extension: bool = False, extension_id: str = "") -> dict | None:
        url = f"{self.graph_url}/{resource}"
        if extension:
            extension_id = (
                extension_id if extension_id else ".".join(reversed(Site.objects.get_current().domain.split(".")))
            )
            url += f"?$expand=Extensions($filter=Id eq '{extension_id}')"
        return self._query_get(url)

    def delete_event_by_resource(self, resource: str) -> dict | None:
        url = f"{self.graph_url}/{resource}"
        return self._query(url, method="DELETE")

    def tentatively_accept_event(self, tenant_id: str, external_id: str) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/tentativelyAccept"
        data = {"sendResponse": False}
        return self._query_create(url, data)

    def decline_event(self, tenant_id: str, external_id: str) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/decline"
        data = {"sendResponse": False}
        return self._query_create(url, data)

    def accept_event(self, tenant_id: str, external_id: str) -> dict | None:
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/accept"
        data = {"sendResponse": False}
        return self._query_create(url, data)

    def cancel_event(self, tenant_id: str, external_id: str) -> dict | None:
        # Only the organizer can cancel an event
        url = f"{self.graph_url}/users/{tenant_id}/events/{external_id}/cancel"
        data = {}
        return self._query_create(url, data)

    def _query_get(self, url: str, raise_error: bool = False) -> dict | None:
        response = self._query(url)
        data = None
        if response.status_code < 400:
            if response.json():
                data = parse(pd.json_normalize(response.json()))
                if data and isinstance(data, list):
                    data = data[0]
        elif raise_error:
            raise ValueError(response, response.json())
        return data

    def _query_get_list(self, url: str, raise_error: bool = False) -> list:
        response = self._query(url)
        datum = []
        if response:
            if (result := response.json()) and (value := result.get("value")):
                datum = parse(pd.json_normalize(value))
                next_url: str | None = result.get("@odata.nextLink")
                while next_url:
                    response = self._query(next_url)
                    if (result := response.json()) and (value := result.get("value")):
                        datum += parse(pd.json_normalize(value))
                        next_url = result.get("@odata.nextLink")
                    else:
                        next_url = None
        elif raise_error:
            raise ValueError(response, response.json())
        return datum

    def _query_create(self, url: str, event_data: dict) -> dict | None:
        response = self._query(url, method="POST", data=json.dumps(event_data))
        data = None
        if response.status_code < 400:
            try:
                if response.json():
                    data = parse(pd.json_normalize(response.json()))
                    if data and isinstance(data, list):
                        data = data[0]
            except requests.exceptions.InvalidJSONError:
                pass
        else:
            raise ValueError(response, response.__dict__)
        return data

    def _query_update(self, url: str, event_data: dict) -> dict | None:
        response = self._query(url, method="PATCH", data=json.dumps(event_data))
        data = None
        if response:
            if response.json():
                data = parse(pd.json_normalize(response.json()))
        else:
            if response.status_code not in [status.HTTP_503_SERVICE_UNAVAILABLE, status.HTTP_409_CONFLICT]:
                raise ValueError(response, response.json())
            else:
                import time

                time.sleep(60)
                if response := self._query(url, method="PATCH", data=json.dumps(data)):
                    if response.json():
                        data = parse(pd.json_normalize(response.json()))
        if data and isinstance(data, list):
            data = data[0]
        return data

    def _query(
        self,
        url: str,
        method: str = "GET",
        data: dict | str | None = None,
        params: dict | None = None,
        access_token: bool = True,
        is_json: bool = True,
    ) -> HttpResponse:
        headers = {"content-type": "application/json" if is_json else "application/x-www-form-urlencoded"}
        if access_token:
            global_preferences = global_preferences_registry.manager()
            headers["Authorization"] = f'Bearer {global_preferences["wbactivity_sync__outlook_sync_access_token"]}'
        if method == "POST":
            response = requests.post(url, data=data, headers=headers)
        elif method == "DELETE":
            response = requests.delete(url, headers=headers)
        elif method == "PATCH":
            response = requests.patch(url, data=data, headers=headers)
        else:
            response = requests.get(url, headers=headers, params=params)
        return response



from typing import Any


def flattened_metadata_lookup(obj: dict, key_string: str = "") -> tuple[str, Any]:
    """
    allows to flatten in nested dictionary which can be used in a lookup query
    """
    if isinstance(obj, dict):
        key_string = key_string + "__" if key_string else key_string
        for k in obj:
            yield from flattened_metadata_lookup(obj[k], key_string + str(k))
    else:
        if isinstance(obj, list):
            yield key_string + "__contains", obj
        else:
            yield key_string, obj


def merge_nested_dict(dct: dict, nested_to_merge: dict):
    """
    allows to merge 2 nested dictionaries
    """
    for k, v in nested_to_merge.items():
        if k in dct:
            if isinstance(dct[k], dict) and isinstance(nested_to_merge[k], dict):  # noqa
                merge_nested_dict(dct[k], nested_to_merge[k])
            else:
                if not isinstance(dct[k], list):
                    dct[k] = [dct[k]]
                dct[k].append(nested_to_merge[k])
        else:
            dct[k] = nested_to_merge[k]


def flattened_dict_into_nested_dict(obj: dict) -> dict:
    """
    allows to nest a flattened dictionary
    """
    nested = {}
    for key, value in obj.items():
        keys = key.split(".")
        dct = {keys.pop(): value}
        while keys:
            dct = {keys.pop(): dct}
        merge_nested_dict(nested, dct)
    return nested



# Generated by Django 5.0.8 on 2024-10-02 07:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0010_alter_account_reference_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="activity",
            name="summary",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AddField(
            model_name="activity",
            name="heat",
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 5.0.10 on 2024-12-06 16:01

import django_fsm
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0016_auto_20241205_1015"),
    ]

    operations = [
        migrations.CreateModel(
            name="Event",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("data", models.JSONField(blank=True, default=dict)),
                ("result", models.JSONField(blank=True, default=dict)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.AlterField(
            model_name="activityparticipant",
            name="participation_status",
            field=django_fsm.FSMField(
                choices=[
                    ("CANCELLED", "Cancelled"),
                    ("MAYBE", "Maybe"),
                    ("ATTENDS", "Attends"),
                    ("NOTRESPONDED", "Not Responded"),
                    ("ATTENDS_DIGITALLY", "Attends Digitally"),
                    ("PENDING_INVITATION", "Pending Invitation"),
                ],
                default="PENDING_INVITATION",
                max_length=50,
                verbose_name="Participation Status",
            ),
        ),
    ]



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

ACCOUNT_MENUITEM = MenuItem(
    label="Accounts",
    endpoint="wbcrm:account-list",
    endpoint_get_parameters={"parent__isnull": True},
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbcrm.view_account"]
    ),
    add=MenuItem(
        label="Create Account",
        endpoint="wbcrm:account-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user), permissions=["wbcrm.add_account"]
        ),
    ),
)



from typing import Optional

from rest_framework.reverse import reverse
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class AccountDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title", pinned="left"),
                dp.Field(key="status", label="Status"),
                dp.Field(
                    key="owner",
                    label="Owner",
                ),
                dp.Field(key="reference_id", label="Reference ID"),
                dp.Field(key="is_terminal_account", label="Terminal Account"),
                dp.Field(key="is_public", label="Public"),
                dp.Field(
                    key="llm",
                    label="LLM Analysis",
                    children=[
                        dp.Field(key="relationship_status", label="Relationship Status"),
                        dp.Field(key="relationship_summary", label="Relationship Summary", show="open"),
                        dp.Field(key="action_plan", label="Action Plan", show="open"),
                    ],
                ),
            ],
            tree=True,
            tree_group_field="title",
            tree_group_field_sortable=True,
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_key="parent",
                    filter_depth=1,
                    # lookup="id_repr",
                    list_endpoint=reverse(
                        "wbcrm:account-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
        )

    def get_instance_display(self) -> Display:
        child_account_section = create_simple_section(
            "child_account_section", "Child Accounts", [["childaccounts"]], "childaccounts", collapsed=True
        )
        account_role_section = create_simple_section(
            "account_role_section", "Account Roles", [["accountroles"]], "accountroles", collapsed=True
        )
        inherited_account_role_section = create_simple_section(
            "inherited_account_role_section",
            "Inherited Account Roles",
            [["inheritedaccountroles"]],
            "inheritedaccountroles",
            collapsed=True,
        )
        return create_simple_display(
            [
                [repeat_field(3, "status")],
                ["title", "title", "reference_id"],
                ["is_active", "is_terminal_account", "is_public"],
                ["parent", "owner", "owner"] if "account_id" not in self.view.kwargs else [repeat_field(3, "owner")],
                [repeat_field(3, "child_account_section")],
                [repeat_field(3, "account_role_section")],
                [repeat_field(3, "inherited_account_role_section")],
            ],
            [child_account_section, account_role_section, inherited_account_role_section],
        )


class AccountRoleAccountDisplayConfig(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display([["role_type", "entry"], ["is_hidden", "authorized_hidden_users"]])

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="role_type", label="Role"),
                dp.Field(key="entry", label="entry"),
                dp.Field(key="is_currently_valid", label="Valid"),
                dp.Field(key="is_hidden", label="Hidden"),
                dp.Field(key="authorized_hidden_users", label="Authorized Hidden Users"),
            ]
        )


class InheritedAccountRoleAccountDisplayConfig(AccountRoleAccountDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="account", label="Account"),
                dp.Field(key="role_type", label="Role"),
                dp.Field(key="entry", label="Entry"),
                dp.Field(key="is_currently_valid", label="Valid"),
                dp.Field(key="is_hidden", label="Hidden"),
                dp.Field(key="authorized_hidden_users", label="Authorized Hidden Users"),
            ]
        )



import json
from unittest.mock import patch

import pytest
from dynamic_preferences.registries import global_preferences_registry
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory

from wbcrm.models.activities import Activity
from wbcrm.synchronization.activity.backends.outlook.backend import OutlookSyncBackend

from .fixtures import TestOutlookSyncFixture, msgraph_fixture


@pytest.mark.django_db
class TestOutlookSyncBackend(TestOutlookSyncFixture):
    backend = OutlookSyncBackend()

    def test_attribute(self):
        assert OutlookSyncBackend.METADATA_KEY == "outlook"

    def test_open(self):
        preferences = global_preferences_registry.manager()
        preferences["wbactivity_sync__outlook_sync_credentials"] = ""
        assert hasattr(self.backend, "msgraph") is False
        self.backend.open()
        assert self.backend.msgraph

    @pytest.mark.parametrize("type_request", [None, "validationToken", "admin_consent"])
    def test_validation_response(self, type_request):
        request1 = APIRequestFactory().get("")
        if type_request:
            request1.GET = request1.GET.copy()
            request1.GET[type_request] = "fake_info"
            assert self.backend._validation_response(request1).content.decode("UTF-8") == "fake_info"
        else:
            assert self.backend._validation_response(request1) is None

    @pytest.mark.parametrize("client_state", [False, True])
    def test_is_inbound_request_valid(
        self, client_state, notification_created_fixture, notification_call_record_fixture
    ):
        api_factory = APIRequestFactory()
        request1 = api_factory.post("", data={})
        request2 = api_factory.post(
            "", data=json.dumps({"value": [notification_call_record_fixture]}), content_type="application/json"
        )
        request3 = api_factory.post(
            "", data=json.dumps({"value": [notification_created_fixture]}), content_type="application/json"
        )
        preferences = global_preferences_registry.manager()
        preferences["wbactivity_sync__outlook_sync_client_state"] = ""
        if client_state:
            preferences["wbactivity_sync__outlook_sync_client_state"] = notification_created_fixture["client_state"]

        assert self.backend._is_inbound_request_valid(request1) is False
        assert self.backend._is_inbound_request_valid(request2) is False
        assert self.backend._is_inbound_request_valid(request3) == client_state

    @pytest.mark.parametrize("event_found, is_organizer", [(False, True), (True, True), (True, False)])
    @patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
    def test_get_events_from_request(
        self,
        mock_msgraph,
        event_found,
        is_organizer,
        notification_created_fixture,
        organizer_event_fixture_parsed,
        invitation_event_fixture_parsed,
    ):
        api_factory = APIRequestFactory()
        request1 = api_factory.post(
            "", data=json.dumps({"value": [notification_created_fixture]}), content_type="application/json"
        )
        self.backend.open()
        self.backend.msgraph = msgraph_fixture()
        if event_found:
            if is_organizer:
                msgraph_fixture.event = organizer_event_fixture_parsed
                msgraph_fixture.tenant_id = "fake_tenant_id"
            else:
                msgraph_fixture.event = invitation_event_fixture_parsed
                msgraph_fixture.event_by_uid = organizer_event_fixture_parsed
        events = self.backend._get_events_from_request(request1)
        self.backend.close()
        assert len(events) == 1
        if event_found:
            assert {"change_type", "resource", "subscription_id"}.issubset(set(events[0].keys()))
            assert events[0].get("id")
        else:
            assert {"change_type", "resource", "subscription_id"} == set(events[0].keys())

    def test_deserialize(self, organizer_event_fixture_parsed, organizer_master_event_fixture_parsed):
        self.backend.open()
        activity_dto0, is_deleted0, user_dto0 = self.backend._deserialize(
            organizer_event_fixture_parsed, include_metadata=False
        )
        activity_dto, is_deleted, user_dto = self.backend._deserialize(organizer_event_fixture_parsed)
        activity_dto2, is_deleted2, user_dto2 = self.backend._deserialize(organizer_master_event_fixture_parsed)
        self.backend.close()

        assert is_deleted0 == is_deleted == is_deleted2 is False
        assert user_dto0 == user_dto == user_dto2
        assert user_dto0.metadata == {"outlook": {"subscription": {"id": None}}}
        assert activity_dto0.metadata == {}
        assert activity_dto.metadata == {
            self.backend.METADATA_KEY: {
                "event_uid": organizer_event_fixture_parsed["uid"],
                "event_id": organizer_event_fixture_parsed["id"],
            }
        }
        assert activity_dto2.metadata == {
            self.backend.METADATA_KEY: {
                "event_uid": organizer_master_event_fixture_parsed["uid"],
                "event_id": organizer_master_event_fixture_parsed["id"],
            }
        }
        assert activity_dto.period
        assert activity_dto.repeat_choice == "NEVER"
        assert activity_dto2.repeat_choice != "NEVER"

    def test_serialize(self, activity_factory):
        activity_dto = activity_factory(preceded_by=None)._build_dto()
        act_dto = activity_factory(
            repeat_choice=Activity.ReoccuranceChoice.DAILY, recurrence_count=3, preceded_by=None
        )._build_dto()
        activity_dict = self.backend._serialize(activity_dto)
        activity_dict1 = self.backend._serialize(activity_dto, created=True)
        act_dict = self.backend._serialize(act_dto)
        act_dict1 = self.backend._serialize(act_dto, created=True)
        keys = {
            "subject",
            "start",
            "end",
            "body",
            "attendees",
            "sensitivity",
            "isReminderOn",
            "reminderMinutesBeforeStart",
            "isAllDay",
            "responseRequested",
            "location",
            "locations",
        }
        assert set(activity_dict.keys()) == set(act_dict.keys()) == set(activity_dict1.keys()) == keys
        assert set(act_dict1.keys()) == keys.union({"recurrence"})

    @pytest.mark.parametrize(
        "metadata, master_event",
        [
            ({}, False),
            ({}, True),
            ({"organizer_resource": "fake_resource"}, False),
            ({"organizer_resource": "fake_resource"}, True),
            ({"event_id": "fake_event_id"}, False),
            ({"event_id": "fake_event_id"}, True),
            ({"event_uid": "fake_event_uid"}, False),
            ({"event_uid": "fake_event_uid"}, True),
            ({"occurrence_resource": "fake_resource"}, False),
            ({"occurrence_resource": "fake_resource"}, True),
            ({"occurrence_id": "fake_occurrence_id"}, False),
            ({"occurrence_id": "fake_occurrence_id"}, True),
        ],
    )
    def test_get_external_event(self, metadata, master_event, activity_factory):
        activity_dto = activity_factory(preceded_by=None)._build_dto()
        activity_dto2 = activity_factory(preceded_by=None, metadata={self.backend.METADATA_KEY: metadata})._build_dto()
        self.backend.open()
        self.backend.msgraph = msgraph_fixture()
        msgraph_fixture.tenant_id = "fake_tenant_id"
        msgraph_fixture.event_by_uid = msgraph_fixture.event = {"id": "event_id"}
        assert (
            self.backend.get_external_event(activity_dto, master_event)
            == self.backend.get_external_event(activity_dto, master_event)
            is None
        )
        if metadata:
            event_result = self.backend.get_external_event(activity_dto2, master_event)
            if master_event or not activity_dto2.is_recurrent:
                if metadata.get("event_uid") or metadata.get("event_id") or metadata.get("organizer_resource"):
                    assert event_result == {"id": "event_id"}
                else:
                    assert event_result is None
            else:
                if metadata.get("occurrence_resource") or metadata.get("occurrence_id"):
                    assert event_result == {"id": "event_id"}
                else:
                    assert event_result is None
        else:
            assert self.backend.get_external_event(activity_dto2, master_event) is None

    def test_get_external_participants(self, activity_factory, person_factory):
        person = person_factory()
        activity_dto = activity_factory(preceded_by=None, participants=(person,))._build_dto()
        activity_dto2 = activity_factory(
            preceded_by=None,
            participants=(person,),
            metadata={self.backend.METADATA_KEY: {"organizer_resource": "fake_resource"}},
        )._build_dto()
        self.backend.open()
        self.backend.msgraph = msgraph_fixture()
        msgraph_fixture.tenant_id = "fake_tenant_id"
        msgraph_fixture.event_by_uid = msgraph_fixture.event = {
            "id": "event_id",
            "attendees": [{"emailAddress": {"address": person._build_dto().email}}],
        }
        assert len(self.backend.get_external_participants(activity_dto, [])) == 0
        assert len(self.backend.get_external_participants(activity_dto2, [])) == 1
        assert len(self.backend.get_external_participants(activity_dto2, activity_dto2.participants)) == 0

    @patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
    def test_is_participant_valid(self, mock_msgraph):
        user = UserFactory(is_superuser=True)
        user2 = UserFactory(is_superuser=True)
        user2.metadata = {self.backend.METADATA_KEY: {"subscription": {"id": "fake_subscription_id"}}}
        user2.save()
        msgraph_fixture._subscription = {"id": "fake_subscription_id"}
        mock_msgraph.return_value = msgraph_fixture()
        assert self.backend._is_participant_valid(user2) is True
        assert self.backend._is_participant_valid(user) is False

    def test_generate_event_metadata(self):
        event = {"id": "fake_id", "uid": "fake_uid"}
        occurrence_event = {"id": "fake_id", "uid": "fake_uid"}
        resource = f"Users/fake_tenant_id/Events/{event['id']}"
        result = {
            "resources": [resource],
            "organizer_resource": resource,
            "event_uid": event["uid"],
            "event_id": event["id"],
        }
        metadata = self.backend._generate_event_metadata("fake_tenant_id", event)
        metadata2 = self.backend._generate_event_metadata("fake_tenant_id", event, occurrence_event)
        assert metadata == {self.backend.METADATA_KEY: result}
        assert metadata2 == {
            self.backend.METADATA_KEY: {**result, **{"occurrence_id": event["id"], "occurrence_resource": resource}}
        }

    def test_get_metadata_from_event(
        self, activity_factory, organizer_event_fixture_parsed, organizer_master_event_fixture_parsed
    ):
        self.backend.open()
        self.backend.msgraph = msgraph_fixture()
        msgraph_fixture.tenant_id = "fake_tenant_id"
        activity_dto = activity_factory(preceded_by=None)._build_dto()
        metadata_list = self.backend._get_metadata_from_event(activity_dto, organizer_event_fixture_parsed)
        # self.backend._get_metadata_from_event()
        resource = f"Users/fake_tenant_id/Events/{organizer_event_fixture_parsed['id']}"
        assert len(metadata_list) == 1
        assert metadata_list[0][0] == activity_dto
        assert metadata_list[0][1] == {
            self.backend.METADATA_KEY: {
                "resources": [resource],
                "organizer_resource": resource,
                "event_uid": organizer_event_fixture_parsed["uid"],
                "event_id": organizer_event_fixture_parsed["id"],
            }
        }

    @patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
    def test_renew_web_hooks(self, mock_msgraph):
        msgraph_fixture._subscription = {"id": "fake_id_new"}
        mock_msgraph.return_value = msgraph_fixture()

        user1 = UserFactory(is_superuser=True)
        user2 = UserFactory(is_superuser=True)
        user2.metadata = {self.backend.METADATA_KEY: {"subscription": {"id": "fake_id"}}}
        user2.save()

        self.backend.renew_web_hooks()
        user1.refresh_from_db()
        user2.refresh_from_db()
        assert user1.metadata == {}
        assert user2.metadata == {self.backend.METADATA_KEY: {"subscription": {"id": "fake_id_new"}}}



import datetime

import pytest
from django.utils import timezone
from psycopg.types.range import DateRange
from selenium.webdriver.common.action_chains import ActionChains
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.authentication.models import User
from wbcore.contrib.authentication.tests.e2e import create_main_company_user
from wbcore.contrib.directory.models import (
    Company,
    EmployerEmployeeRelationship,
    Person,
)
from wbcore.contrib.directory.serializers import (
    ClientManagerModelSerializer,
    CompanyModelSerializer,
    PersonModelSerializer,
)
from wbcore.contrib.directory.tests.e2e import (
    create_new_cmr_instance,
    create_new_company_instance,
    create_new_person_instance,
    set_up_companies,
    set_up_persons,
)
from wbcore.contrib.directory.viewsets.display import CMR_Color
from wbcore.test import (
    click_button_by_label,
    click_close_all_widgets,
    click_new_button,
    delete_list_entry,
    does_background_color_match,
    edit_list_instance,
    find_element_by_text,
    is_counter_as_expected,
    is_error_visible,
    is_string_not_visible,
    is_tag_not_visible,
    is_tag_visible,
    is_text_visible,
    navigate_to_filter,
    open_menu_item,
    select_async_filter,
    select_column_filter,
    select_filter,
    set_up,
)

from wbcrm.factories import ActivityFactory
from wbcrm.models import Activity
from wbcrm.serializers import ActivityModelSerializer
from wbcrm.tests.e2e import create_new_activity_instance, set_up_activities

USER_PASSWORD = "User_Password"


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
class TestCRMFilters:
    def test_entry_filter(self, live_server, selenium):
        """
        Testing the "Entry"-Filter.
        """

        # Creating a test user
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        # Creating three test companies with different last activities.
        companies = set_up_companies()
        now = timezone.now()
        period_a = DateRange(now - datetime.timedelta(days=15, hours=2), now - datetime.timedelta(days=15))
        period_b = DateRange(now - datetime.timedelta(days=60, hours=2), now - datetime.timedelta(days=60))
        ActivityFactory(period=period_a, companies=[companies[0]])
        ActivityFactory(period=period_b, companies=[companies[1], companies[2]])

        set_up(selenium, live_server, user.email, USER_PASSWORD)
        navigate_to_filter(selenium, "CRM", "Companies")

        # And the beginning all three companies should be visible, since no filter is set.
        assert is_counter_as_expected(selenium, Company.objects.count())

        # With no filters selected the there should be a message with the hint "No filter criteria yet"
        assert is_text_visible(selenium, "No filter criteria yet")

        # Select a "No Activity" - filter. Filtering for all companies, that had no activity in the last month.
        select_filter(selenium, "No Activity:", "Last Month")

        # The "No filter criteria yet" text should now be gone and instead there should be a "filter-tag" with "Last Month".
        assert is_string_not_visible(selenium, "No filter criteria yet")
        assert is_tag_visible(selenium, "Last Month")

        # Only two companies should be listed, since one company had an activity withing the last month.
        assert is_text_visible(selenium, "Company B")
        assert is_text_visible(selenium, "Company C")
        assert is_string_not_visible(selenium, "Company A")
        assert is_counter_as_expected(selenium, 2)

        # Selecting an additional column filter to filter by company type.
        select_column_filter(selenium, "Type", "Type A")

        # After selecting the additional filter there should also be an additional filter tag corresponding to the filter and the number of companies should be one.
        assert is_tag_visible(selenium, "Last Month")
        assert is_tag_visible(selenium, "Type A")
        assert is_counter_as_expected(selenium, 1)
        assert is_text_visible(selenium, "Company B")
        assert is_string_not_visible(selenium, "Company C")

        # Resetting the filters
        click_button_by_label(selenium, "Clear all filters")
        assert is_text_visible(selenium, "No filter criteria yet")
        assert is_tag_not_visible(selenium, "Type A")
        assert is_counter_as_expected(selenium, Company.objects.count())

        # Close all widgets and switch to the person list
        click_close_all_widgets(selenium)
        set_up_persons()
        navigate_to_filter(selenium, "Persons")

        assert is_counter_as_expected(selenium, Person.objects.count())
        assert is_text_visible(selenium, "No filter criteria yet")

        # Select a "Specializations" - filter. Filtering for all persons with "Specialization C".
        select_async_filter(selenium, "Specializations:", "Specialization C")
        assert is_tag_visible(selenium, "Specialization C")
        assert is_string_not_visible(selenium, "No filter criteria yet")

        # Only two persons should be listed
        assert is_string_not_visible(selenium, "Henry Kalb")
        assert is_text_visible(selenium, "Konrad Zuse")
        assert is_text_visible(selenium, "Ada Lovelace")
        filter_count = Person.objects.filter(specializations__title="Specialization C").count()
        assert is_counter_as_expected(selenium, filter_count)

        # Selecting an additional column filter to filter by person status.
        select_column_filter(selenium, "Status", "Status C")

        # After selecting the additional filter there should also be an additional filter tag corresponding to the filter and the number of persons should be one.
        assert is_tag_visible(selenium, "Specialization C")
        assert is_tag_visible(selenium, "Status C")
        assert is_string_not_visible(selenium, "Konrad Zuse")
        assert is_text_visible(selenium, "Ada Lovelace")
        filter_count = EmployerEmployeeRelationship.objects.filter(
            employer__in=Company.objects.filter(customer_status__title="Status C"),
            employee__in=Person.objects.filter(specializations__title="Specialization C"),
            primary=True,
        )
        assert is_counter_as_expected(selenium, filter_count)

    def test_activity_filter(self, live_server, selenium):
        """
        Testing the "Type" and "Status" - Filter
        """
        # Creating a test user and setting up selenium
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        set_up_activities(user.profile)
        set_up(selenium, live_server, user.email, USER_PASSWORD)

        navigate_to_filter(selenium, "CRM", "Activities")

        # And the beginning all three activities should be visible, since no filter is set.
        filter_count = Activity.objects.all().count()
        assert is_counter_as_expected(selenium, filter_count)

        # With no filters selected the there should be a message with the hint "No filter criteria yet"
        assert is_text_visible(selenium, "No filter criteria yet")

        # Selecting a column filter to filter by activity type.
        select_column_filter(selenium, "Type", "Meeting")

        assert is_string_not_visible(selenium, "No filter criteria yet")
        assert is_tag_visible(selenium, "Meeting")

        # After filtering for the type there should be only two activities displayed
        filter_count = Activity.objects.filter(type__title="Meeting").count()
        assert is_counter_as_expected(selenium, filter_count)
        assert is_text_visible(selenium, "Activity A")
        assert is_text_visible(selenium, "Activity B")
        assert is_string_not_visible(selenium, "Activity C")

        # Select a "Status" - filter. Filtering for all activities, that have the status "Finished".
        select_filter(selenium, "Status:", "Finished")

        assert is_string_not_visible(selenium, "Activity A")
        assert is_text_visible(selenium, "Activity B")
        filter_count = Activity.objects.filter(type__title="Meeting", status=Activity.Status.FINISHED).count()
        assert is_counter_as_expected(selenium, filter_count)


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
class TestCRMActivities:
    def test_create_edit_delete_activity(self, live_server, selenium):
        """
        Creates, edits and deletes an activity.
        """
        # Creating a test user and setting up selenium
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        actions = ActionChains(selenium, 1000)
        set_up(selenium, live_server, user.email, USER_PASSWORD)

        # # -----> CREATE <----- #
        activity_a = create_new_activity_instance(selenium, ["title", "type"], "Type A", "Activity A", False)
        open_menu_item(selenium, "Activities", perform_mouse_move=True)
        assert is_text_visible(selenium, activity_a.title)
        assert is_counter_as_expected(selenium, Activity.objects.count())

        click_new_button(selenium)
        assert is_text_visible(selenium, "New Activity")

        # Trying to create an activity without filling out anything -> We expect an error to be thrown.
        click_button_by_label(selenium, "Save and close")
        assert is_error_visible(selenium)

        activity_b = create_new_activity_instance(selenium, ["title", "type"], None, "Activity B")
        assert is_text_visible(selenium, activity_b.title)
        assert is_counter_as_expected(selenium, Activity.objects.count())

        # -----> Edit <----- #

        edit_list_instance(selenium, actions, activity_a, ActivityModelSerializer(activity_a), {"title": "A Activity"})

        assert is_string_not_visible(selenium, "Activity A")
        assert is_text_visible(selenium, "A Activity")
        assert is_counter_as_expected(selenium, Activity.objects.count())

        # -----> Delete <----- #
        delete_list_entry(selenium, actions, activity_b.title)
        assert is_counter_as_expected(selenium, Activity.objects.count())
        assert is_string_not_visible(selenium, "Activity A")
        assert is_string_not_visible(selenium, "Activity B")


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
class TestCRMEntries:
    def test_create_edit_delete_company(self, live_server, selenium):
        """
        Creating a new company from the company list view. After creating the company it should be displayed in the company list view.
        """
        # Creating a test user
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        actions = ActionChains(selenium, 1000)
        set_up(selenium, live_server, user.email, USER_PASSWORD)

        # -----> CREATE <----- #
        company_a = create_new_company_instance(
            selenium, ["name", "customer_status"], "Company A", "Test Status", False
        )
        open_menu_item(selenium, "Companies", perform_mouse_move=True)
        assert is_text_visible(selenium, company_a.name)
        assert is_counter_as_expected(selenium, Company.objects.count())

        click_new_button(selenium)
        assert is_text_visible(selenium, "New Company")

        # Trying to create an activity without filling out anything -> We expect an error to be thrown.
        click_button_by_label(selenium, "Save and close")
        assert is_error_visible(selenium)
        company_b = create_new_company_instance(selenium, ["name", "customer_status"], "Company B")

        assert is_text_visible(selenium, company_b.name)
        assert is_counter_as_expected(selenium, Company.objects.count())

        # -----> Edit <----- #

        edit_list_instance(selenium, actions, company_a.name, CompanyModelSerializer(company_a), {"name": "A Company"})
        assert is_string_not_visible(selenium, "Company A")
        assert is_text_visible(selenium, "A Company")
        assert is_counter_as_expected(selenium, Company.objects.count())

        # -----> Delete <----- #

        delete_list_entry(selenium, actions, company_b.name)
        assert is_counter_as_expected(selenium, Company.objects.count())
        assert is_text_visible(selenium, "A Company")
        assert is_string_not_visible(selenium, "Company A")
        assert is_string_not_visible(selenium, "Company B")

    def test_create_edit_delete_person(self, live_server, selenium):
        """
        Creating a new person from the person list view. After creating the person it should be displayed in the person list view.
        """
        # Creating a test user
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        actions = ActionChains(selenium, 1000)
        set_up(selenium, live_server, user.email, USER_PASSWORD)

        # -----> CREATE <----- #
        person_a = create_new_person_instance(
            selenium, ["first_name", "last_name", "prefix"], "Max", "Mustermann", "Mr.", False
        )
        open_menu_item(selenium, "Persons", perform_mouse_move=True)
        assert is_text_visible(selenium, f"{person_a.first_name} {person_a.last_name}")
        assert is_counter_as_expected(selenium, Person.objects.count())

        click_new_button(selenium)
        assert is_text_visible(selenium, "New Person")

        # Trying to create an activity without filling out anything -> We expect an error to be thrown.
        click_button_by_label(selenium, "Save and close")
        assert is_error_visible(selenium)
        person_b = create_new_person_instance(
            selenium, ["first_name", "last_name", "prefix"], "Maike", "Musterfrau", "Mrs.", False
        )

        assert is_text_visible(selenium, f"{person_b.first_name} {person_b.last_name}")
        assert is_counter_as_expected(selenium, Person.objects.count())

        # -----> Edit <----- #

        edit_list_instance(
            selenium,
            actions,
            f"{person_a.first_name} {person_a.last_name}",
            PersonModelSerializer(person_a),
            {"first_name": "Maxissimus"},
        )
        assert is_string_not_visible(selenium, "Max Mustermann")
        assert is_text_visible(selenium, "Maxissimus Mustermann")
        assert is_counter_as_expected(selenium, Person.objects.count())

        # -----> Delete <----- #

        delete_list_entry(selenium, actions, f"{person_b.first_name} {person_b.last_name}")
        assert is_counter_as_expected(selenium, Person.objects.count())
        assert is_text_visible(selenium, "Maxissimus Mustermann")
        assert is_string_not_visible(selenium, "Max Mustermann")
        assert is_string_not_visible(selenium, "Maike Mustermann")


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
# This test class is currently bugged because of frontend issues.
class TestCRMRelationShipManagement:
    def test_create_edit_and_delete_relationship(self, live_server, selenium):
        """
        Creates, edits and deletes a Customer-Employer-Relationship.
        """
        user: User = create_main_company_user(USER_PASSWORD)
        actions = ActionChains(selenium, 1000)
        set_up(selenium, live_server, user.email, USER_PASSWORD)
        set_up_persons()

        # -----> CREATE <----- #
        cmr = create_new_cmr_instance(selenium, ["client", "relationship_manager", "primary"], True, False)
        open_menu_item(selenium, "Client Manager Relationships", perform_mouse_move=True)
        cmr_element = find_element_by_text(selenium, cmr.client.computed_str)

        assert is_counter_as_expected(selenium, 1)
        assert is_text_visible(selenium, cmr.client.computed_str)
        assert is_text_visible(selenium, cmr.relationship_manager.computed_str)
        assert does_background_color_match(cmr_element, CMR_Color.DRAFT.value)

        # # -----> Edit <----- #

        edit_list_instance(
            selenium,
            actions,
            cmr.client.computed_str,
            ClientManagerModelSerializer(cmr),
            {"relationship_manager": user.profile.computed_str},
        )
        assert is_string_not_visible(selenium, cmr.relationship_manager.computed_str)
        assert is_text_visible(selenium, user.profile.computed_str)
        assert is_counter_as_expected(selenium, Person.objects.count())

        # # -----> Approve <----- #
        # TODO Implement approve
        # -----> Delete <----- #
        # TODO Implement delete



from datetime import datetime, timedelta
from typing import Dict, List
from zoneinfo import ZoneInfo

from django.conf import settings
from django.db import models
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.directory.models import EmailContact, Person

from wbcrm.models import Activity, ActivityParticipant
from wbcrm.synchronization.activity.preferences import (
    can_synchronize_activity_description,
)

from .typing_informations import GoogleEventType, TimeInfo

global_preferences = global_preferences_registry.manager()


class GoogleSyncUtils:
    @classmethod
    def convert_event_visibility_to_activity_visibility(cls, event_visiblity: str):
        if event_visiblity == "public":
            return CalendarItem.Visibility.PUBLIC
        return CalendarItem.Visibility.PRIVATE

    @classmethod
    def convert_activity_participants_to_attendees(cls, activity: Activity, event: GoogleEventType | None = None):
        """
        A method for converting the participants of a "Workbench" activity to a list of dictionaries, that can be used to create "Google Event" attendees.

        :param activity:    The activity from which the participants needs to be converted.
        :return:            The list with the dictionaries of attendees. Can be empty if no participant is in the original activity.
        """
        attendees_list = []
        event_attendees = cls.get_or_create_participants(event, activity.creator) if event else []

        event_attendees = [Person.all_objects.get(id=person_dict["person_id"]) for person_dict in event_attendees]
        can_sync_external_participants: bool = global_preferences["wbactivity_sync__sync_external_participants"]
        allowed_participants: models.QuerySet["ActivityParticipant"] = activity.activity_participants.filter(
            participant__in=event_attendees,
        ).union(activity.activity_participants.filter(participant__id__in=Person.objects.filter_only_internal()))
        activity_participants: models.QuerySet["ActivityParticipant"] = (
            activity.activity_participants.all() if can_sync_external_participants else allowed_participants
        )
        for activity_participant in activity_participants:
            participant = activity_participant.participant
            status = cls.convert_participant_status_to_attendee_status(activity_participant.participation_status)
            attendees_list.append(
                {
                    "displayName": participant.computed_str,
                    "email": str(participant.primary_email_contact()),
                    "responseStatus": status,
                }
            )
        return attendees_list

    @classmethod
    def convert_activity_to_event(cls, activity: Activity, created=False):
        """
        Converts a "Workbench" activity into a dict, that can be used to create a "Google Event" instance.

        :param activity:    A "Workbench" activity. The activity that is converted to a dict.
        :param created:     A boolean. This should be True if the activity is created, in every other case it should be False. Per default it is False.
        :returns:           A dictionary which can be used to create a google event.
        """
        # If the activity instance is to be created (e.g. pre_save when creating the activity) we cannot interact with the activity_participants.
        participants_list = cls.convert_activity_participants_to_attendees(activity=activity)
        timezone = ZoneInfo(settings.TIME_ZONE)
        recurrence = []
        if not created:
            if (google_backend := activity.metadata.get("google_backend")) and (event := google_backend.get("event")):
                recurrence = event.get("recurrence", [])
            elif wb_recurrence := activity.metadata.get("recurrence"):
                recurrence = [wb_recurrence]
        else:
            if wb_recurrence := activity.metadata.get("recurrence"):
                recurrence = [wb_recurrence]
            else:
                recurrence = []
        event_body = {
            "summary": activity.title,
            "creator": str(activity.creator.primary_email_contact()) if activity.creator else "",
            "organizer": str(activity.assigned_to.primary_email_contact()) if activity.assigned_to else "",
            "attendees": participants_list,
            "description": activity.description if can_synchronize_activity_description else "",
            "start": {
                "dateTime": activity.period.lower.astimezone(timezone).isoformat(),  # type: ignore
                "timeZone": settings.TIME_ZONE,
            },
            "end": {
                "dateTime": activity.period.upper.astimezone(timezone).isoformat(),  # type: ignore
                "timeZone": settings.TIME_ZONE,
            },
            "recurrence": recurrence,
            "location": activity.location,
            "visibility": cls.convert_activity_visibility_to_event_visibility(activity.visibility),
            "reminders": {
                "useDefault": False,
                "overrides": [
                    {
                        "method": "email",
                        "minutes": Activity.ReminderChoice.get_minutes_correspondance(activity.reminder_choice),
                    }
                ],
            },
        }
        return event_body

    @classmethod
    def get_start_and_end(cls, event: GoogleEventType):
        """
        Converts the google start & end times to a datetime format.

        A google event dict contains either a dateTime-key or a date-key. If a date-key is present, this indicates that the event is an all day event.

        :param event: A google event dictionary from which the start & and end time will be extracted and converted.

        :note: If a google event takes place all day, the start date and end date will not be the same date.
        For example, an event that will take place on the 01.06.2020 will have "2020-06-01" as the start date value and "2020-06-02" as the end date value.
        A workbench activity will work with "datetime" values, even when the activity is an all-day activity.
        Therefor the start date will be converted in a "datetime" object like "2020-06-01-00-00-00" and the end date will be converted to "2020-06-01-23-59-59"
        """
        event_start, event_end = None, None
        start, end = event["start"], event["end"]
        if start_datetime := start.get("dateTime"):
            event_start = datetime.strptime(start_datetime, "%Y-%m-%dT%H:%M:%S%z")
        else:
            event_start = datetime.strptime(start["date"], "%Y-%m-%d")
        if end_datetime := end.get("dateTime"):
            event_end = datetime.strptime(end_datetime, "%Y-%m-%dT%H:%M:%S%z")
        else:
            event_end = datetime.strptime(end["date"], "%Y-%m-%d") - timedelta(seconds=1)
        return event_start, event_end

    @classmethod
    def add_instance_metadata(
        cls, parent_occurrence: Activity, google_event_items: List[GoogleEventType], new_metadata: Dict, created=False
    ) -> None:
        """
        Adds the information of google event instances to the corresponding activities metadata field.

        :param parent_occurrence: The "Workbench" parent activity. It is used to retrieve the child activities.
        :google_event_items: The corresponding google event instances.
        :new_metadata: The newly created metadata for the parent_occurrence
        """

        def compare_instances_and_activity(activity: Activity, is_parent=False):
            for google_child in google_event_items:
                if created:
                    google_start, _ = cls.get_start_and_end(google_child)
                    activity_start = activity.period.lower  # type: ignore
                else:
                    original_start_dict: TimeInfo = google_child.get("originalStartTime", {})
                    activity_start_dict: Dict = (
                        activity.metadata.get("google_backend", {}).get("instance", {}).get("originalStartTime", {})
                    )
                    google_start = original_start_dict.get("dateTime", original_start_dict.get("date"))
                    activity_start = activity_start_dict.get("dateTime", original_start_dict.get("date"))
                if google_start == activity_start:
                    if is_parent:
                        new_metadata["google_backend"] |= {"instance": google_child}
                        external_id = new_metadata["google_backend"]["event"].get("id")
                        Activity.objects.filter(id=activity.id).update(external_id=external_id, metadata=new_metadata)
                    else:
                        metadata = activity.metadata | {"google_backend": {"instance": google_child}}
                        Activity.objects.filter(id=activity.id).update(
                            external_id=google_child["id"], metadata=metadata
                        )
                    google_event_items.remove(google_child)
                    break

        compare_instances_and_activity(parent_occurrence, True)
        child_activities = Activity.objects.filter(parent_occurrence=parent_occurrence)
        for wb_child in child_activities:
            compare_instances_and_activity(wb_child)

    @classmethod
    def get_or_create_participants(cls, event: GoogleEventType, creator: Person | None):
        """
        Converts the participants of an event into person objects by using the email address of the participants to search for the corresponding person entries in the "Workbench" database.
        If no person with the right email address is found, a new person entry will automatically be created with the email address or, if possible, with the displayed name as last name.

        :param event:       A google event body.
        :param creator:     The creator of the event.

        :return:            A list of dicts with the following structure: {"person_id": int, "status":str}. Can be empty if the event has no attendees.
        """

        participants = (
            [{"person_id": creator.id, "status": ActivityParticipant.ParticipationStatus.ATTENDS}] if creator else []
        )
        if attendees := event.get("attendees"):
            for attendee in attendees:
                mail = attendee.get("email")
                mail_query = Person.objects.filter(
                    id__in=EmailContact.objects.filter(address__iexact=mail, primary=True).values_list(
                        "entry_id", flat=True
                    )
                )
                if mail and mail_query.exists():
                    person = mail_query.first()
                else:
                    if display_name := attendee.get("displayName"):
                        person = Person.objects.create(last_name=display_name, is_draft_entry=True)
                    else:
                        person = Person.objects.create(last_name=mail, is_draft_entry=True)
                    EmailContact.objects.create(entry=person, address=mail, primary=True)
                if not creator or person.id != creator.id:
                    person_obj = {
                        "person_id": person.id,
                        "status": cls.convert_attendee_status_to_participant_status(attendee.get("responseStatus")),
                    }
                    participants.append(person_obj)

        return participants
